(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = global || self, global.AbstractNestedRouter = factory());
}(this, (function () { 'use strict';

  var defaultExport = /*@__PURE__*/(function (Error) {
    function defaultExport(route, path) {
      var message = "Unreachable '" + (route !== '/' ? route.replace(/\/$/, '') : route) + "', segment '" + path + "' is not defined";
      Error.call(this, message);
      this.message = message;
      this.route = route;
      this.path = path;
    }

    if ( Error ) defaultExport.__proto__ = Error;
    defaultExport.prototype = Object.create( Error && Error.prototype );
    defaultExport.prototype.constructor = defaultExport;

    return defaultExport;
  }(Error));

  function buildMatcher(path, parent) {
    var regex;

    var _isSplat;

    var _priority = -100;

    var keys = [];
    regex = path.replace(/[-$.]/g, '\\$&').replace(/\(/g, '(?:').replace(/\)/g, ')?').replace(/([:*]\w+)(?:<([^<>]+?)>)?/g, function (_, key, expr) {
      keys.push(key.substr(1));

      if (key.charAt() === ':') {
        _priority += 100;
        return ("((?!#)" + (expr || '[^#/]+?') + ")");
      }

      _isSplat = true;
      _priority += 500;
      return ("((?!#)" + (expr || '[^#]+?') + ")");
    });

    try {
      regex = new RegExp(("^" + regex + "$"));
    } catch (e) {
      throw new TypeError(("Invalid route expression, given '" + parent + "'"));
    }

    var _hashed = path.includes('#') ? 0.5 : 1;

    var _depth = path.length * _priority * _hashed;

    return {
      keys: keys,
      regex: regex,
      _depth: _depth,
      _isSplat: _isSplat
    };
  }
  var PathMatcher = function PathMatcher(path, parent) {
    var ref = buildMatcher(path, parent);
    var keys = ref.keys;
    var regex = ref.regex;
    var _depth = ref._depth;
    var _isSplat = ref._isSplat;
    return {
      _isSplat: _isSplat,
      _depth: _depth,
      match: function (value) {
        var matches = value.match(regex);

        if (matches) {
          return keys.reduce(function (prev, cur, i) {
            prev[cur] = typeof matches[i + 1] === 'string' ? decodeURIComponent(matches[i + 1]) : null;
            return prev;
          }, {});
        }
      }
    };
  };

  PathMatcher.push = function push (key, prev, leaf, parent) {
    var root = prev[key] || (prev[key] = {});

    if (!root.pattern) {
      root.pattern = new PathMatcher(key, parent);
      root.route = (leaf || '').replace(/\/$/, '') || '/';
    }

    prev.keys = prev.keys || [];

    if (!prev.keys.includes(key)) {
      prev.keys.push(key);
      PathMatcher.sort(prev);
    }

    return root;
  };

  PathMatcher.sort = function sort (root) {
    root.keys.sort(function (a, b) {
      return root[a].pattern._depth - root[b].pattern._depth;
    });
  };

  function merge(path, parent) {
    return ("" + (parent && parent !== '/' ? parent : '') + (path || ''));
  }
  function walk(path, cb) {
    var matches = path.match(/<[^<>]*\/[^<>]*>/);

    if (matches) {
      throw new TypeError(("RegExp cannot contain slashes, given '" + matches + "'"));
    }

    var parts = path.split(/(?=\/|#)/);
    var root = [];

    if (parts[0] !== '/') {
      parts.unshift('/');
    }

    parts.some(function (x, i) {
      var parent = root.slice(1).concat(x).join('') || null;
      var segment = parts.slice(i + 1).join('') || null;
      var retval = cb(x, parent, segment ? ("" + (x !== '/' ? x : '') + segment) : null);
      root.push(x);
      return retval;
    });
  }
  function reduce(key, root, _seen) {
    var parent = root.refs;
    var params = {};
    var out = [];
    var splat;
    walk(key, function (x, leaf, extra) {
      if (!root.keys) {
        throw new defaultExport(key, x);
      }

      var found;
      root.keys.some(function (k) {
        if (_seen.includes(k)) { return false; }
        var ref = root[k].pattern;
        var match = ref.match;
        var _isSplat = ref._isSplat;
        var matches = match(_isSplat ? extra || x : x);

        if (matches) {
          Object.assign(params, matches);

          if (root[k].route) {
            var routeInfo = Object.assign({}, root[k].info); // properly handle exact-routes!

            var hasMatch = false;

            if (routeInfo.exact) {
              hasMatch = extra === null;
            } else {
              hasMatch = !(x && leaf === null) || x === leaf || _isSplat || !extra;
            }

            routeInfo.matches = hasMatch;
            routeInfo.params = Object.assign({}, params);
            routeInfo.route = root[k].route;
            routeInfo.path = _isSplat && extra || leaf || x;

            if (parent[routeInfo.route]) {
              routeInfo.key = parent[routeInfo.route];

              if (routeInfo.path.substr(-1) === '/') {
                routeInfo.key = parent[((routeInfo.route) + "/")] || routeInfo.key;
              }
            }

            out.push(routeInfo);
          }

          if (extra === null && !root[k].keys) {
            return true;
          }

          if (k !== '/') { _seen.push(k); }
          splat = _isSplat;
          root = root[k];
          found = true;
          return true;
        }

        return false;
      });

      if (!(found || root.keys.some(function (k) { return root[k].pattern.match(x); }))) {
        throw new defaultExport(key, x);
      }

      return splat || !found;
    });
    return out;
  }
  function find(path, routes, retries) {
    var get = reduce.bind(null, path, routes);
    var set = [];

    while (retries > 0) {
      retries -= 1;

      try {
        return get(set);
      } catch (e) {
        if (retries > 0) {
          return get(set);
        }

        throw e;
      }
    }
  }
  function add(path, routes, parent, routeInfo) {
    var fullpath = merge(path, parent);
    var params = Object.assign({}, routeInfo);
    var root = routes;
    walk(fullpath, function (x, leaf) {
      root = PathMatcher.push(x, root, leaf, fullpath);

      if (x !== '/') {
        root.info = Object.assign({}, params, root.info);
      }
    });

    if (params.key && !routes.refs[params.key]) {
      routes.refs[params.key] = fullpath;
      routes.refs[fullpath] = params.key;
    }

    root.info = Object.assign({}, root.info, routeInfo);

    if (fullpath !== '/' && fullpath.substr(-1) === '/') {
      var prop = fullpath.replace(/(?<=.)\/$/, '');
      routes.refs[prop] = routes.refs[prop] || params.key;
    }

    return fullpath;
  }
  function rm(path, routes, parent) {
    var fullpath = merge(path, parent);
    var root = routes;
    var leaf = null;
    var key = null;
    walk(fullpath, function (x) {
      if (!root) {
        leaf = null;
        return true;
      }

      if (!root.keys) {
        throw new defaultExport(path, x);
      }

      key = x;
      leaf = root;
      root = root[key];
    });

    if (!(leaf && key)) {
      throw new defaultExport(path, key);
    }

    if (leaf === routes) {
      leaf = routes['/'];
    }

    if (leaf.route !== key) {
      var offset = leaf.keys.indexOf(key);

      if (offset === -1) {
        throw new defaultExport(path, key);
      }

      leaf.keys.splice(offset, 1);
      PathMatcher.sort(leaf);
      delete leaf[key];
    } // nested routes are upgradeable, so keep original info...


    if (root.route === leaf.route && (!root.info || root.info.key === leaf.info.key)) { delete leaf.info; }
  }

  var Router = function Router() {
    var routes = {
      refs: {}
    };
    var stack = [];
    return {
      resolve: function (path, cb) {
        var url = path.split('?')[0];
        var seen = [];
        walk(url, function (x, leaf, extra) {
          try {
            cb(null, find(leaf, routes, 1).filter(function (r) {
              if (!seen.includes(r.path)) {
                seen.push(r.path);
                return true;
              }

              return false;
            }));
          } catch (e) {
            cb(e, []);
          }
        });
      },
      mount: function (path, cb) {
        if (path !== '/') {
          stack.push(path);
        }

        cb();
        stack.pop();
      },
      find: function (path, retries) { return find(path, routes, retries === true ? 2 : retries || 1); },
      add: function (path, routeInfo) { return add(path, routes, stack.join(''), routeInfo); },
      rm: function (path) { return rm(path, routes, stack.join('')); }
    };
  };

  Router.matches = function matches (uri, path) {
    return buildMatcher(uri, path).regex.test(path);
  };

  return Router;

})));
